package com.flyfish.framework.domain.base;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.flyfish.framework.query.Queries;
import com.flyfish.framework.query.Query;
import com.flyfish.framework.query.QueryDefinition;
import com.flyfish.framework.utils.CopyUtils;
import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;
import org.springframework.data.domain.Example;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.repository.core.EntityInformation;

import java.lang.reflect.ParameterizedType;
import java.util.List;
import java.util.Optional;

import static org.apache.commons.lang3.ObjectUtils.defaultIfNull;

/**
 * 基本的查询实体
 *
 * @author Mr.Wang
 */
@Slf4j
public class BaseQo<T extends Domain> implements Qo<T> {

    protected Pageable pageable;

    protected List<T> result;

    protected IUser user;

    protected List<String> fields;

    @Getter
    @Setter
    protected boolean fetchRef;

    public Qo<T> accept(List<T> result, Pageable pageable) {
        this.pageable = pageable;
        this.result = result;
        return this;
    }

    public Qo<T> accept(Pageable pageable) {
        this.pageable = pageable;
        return this;
    }

    @Override
    public Pageable getPageable() {
        return pageable;
    }

    @JsonIgnore
    @Override
    public void setPageable(Pageable pageable) {
        this.pageable = pageable;
    }

    /**
     * 获取当前用户
     *
     * @return 结果
     */
    @Override
    public IUser getUser() {
        return user;
    }

    /**
     * 当前用户
     *
     * @param user 用户数据
     */
    @Override
    public void setUser(IUser user) {
        this.user = user;
    }

    @Override
    public List<T> getResult() {
        return result;
    }

    /**
     * 设置结果集，用于单例部署时快速封装
     *
     * @param result 查询结果集
     */
    @JsonIgnore
    @Override
    public void setResult(List<T> result) {
        this.result = result;
    }

    /**
     * 定义查询规则的方法
     *
     * @return 包含匹配规则的example
     */
    @JsonIgnore
    @Override
    public Example<T> getExample() {
        return null;
    }

    /**
     * 获取data-mongo的对象Criteria
     *
     * @return 结果
     */
    @Override
    public <C> Optional<C> getQuery(EntityInformation<T, String> entityInformation) {
        Query query = Query.empty();
        QueryDefinition definition = queryBuilder();
        if (null != definition) {
            // 查询定义存在时，使用查询定义
            query.setDefinition(definition);
        } else {
            // 不存在查询定义，取得example，若不存在example，则使用probe
            query.setExample(defaultIfNull(getExample(), getProbe()));
        }
        query.setSort(sorts());
        query.setPageable(getPageable());
        query.setFields(getFields());
        return Optional.of(Queries.convert(query, entityInformation));
    }

    /**
     * 获取查询的字段
     *
     * @return 结果
     */
    @Override
    public List<String> getFields() {
        return fields;
    }

    /**
     * 内置的builder，重写后自动合并结果
     *
     * @return 结果
     */
    public QueryDefinition queryBuilder() {
        return null;
    }

    /**
     * 让值全部包含在Pojo里
     *
     * @return 结果
     */
    @Override
    @SuppressWarnings("unchecked")
    public Class<T> pojoType() {
        ParameterizedType type = (ParameterizedType) this.getClass().getGenericSuperclass();
        return (Class<T>) type.getActualTypeArguments()[0];
    }


    @Override
    public Sort sorts() {
        return Sort.by(Sort.Order.desc("modifyTime"));
    }

    /**
     * 判断当前查询是否为空
     *
     * @return 结果
     */
    @Override
    public boolean isEmpty() {
        QueryDefinition definition = queryBuilder();
        Example<T> example = getExample();
        return null == example && (null == definition || definition.isEmpty());
    }

    private Example<T> getProbe() {
        Class<T> type = pojoType();
        if (null != type && !Object.class.equals(type)) {
            try {
                T pojo = CopyUtils.copyQueryProps(this, type.newInstance());
                return Example.of(pojo);
            } catch (InstantiationException | IllegalAccessException e) {
                log.warn("尝试实例化bean失败！", e);
            }
        }
        return null;
    }
}
