package com.flyfish.framework.mongodb.repository.impl;

import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.domain.base.Qo;
import com.flyfish.framework.repository.DefaultReactiveRepository;
import lombok.Getter;
import lombok.NonNull;
import org.apache.commons.lang3.StringUtils;
import org.springframework.dao.IncorrectResultSizeDataAccessException;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.mongodb.core.ReactiveMongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.repository.query.MongoEntityInformation;
import org.springframework.data.mongodb.repository.support.SimpleReactiveMongoRepository;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.List;
import java.util.Optional;

/**
 * 默认的异步仓库实现
 *
 * @param <T> 泛型
 */
public class DefaultReactiveRepositoryImpl<T extends Domain> extends SimpleReactiveMongoRepository<T, String>
        implements DefaultReactiveRepository<T> {

    @Getter
    private final MongoEntityInformation<T, String> entityInformation;

    private final ReactiveMongoOperations mongoOperations;

    public DefaultReactiveRepositoryImpl(@NonNull MongoEntityInformation<T, String> entityInformation,
                                         @NonNull ReactiveMongoOperations mongoOperations) {
        super(entityInformation, mongoOperations);
        this.mongoOperations = mongoOperations;
        this.entityInformation = entityInformation;
    }

    /**
     * 通过名称查找一个
     *
     * @param name 名称
     * @return 结果
     */
    @Override
    public Mono<T> findByName(String name) {
        if (StringUtils.isNotBlank(name)) {
            return mongoOperations.findOne(Query.query(Criteria.where("name").is(name)),
                    entityInformation.getJavaType(), entityInformation.getCollectionName());
        }
        return Mono.empty();
    }

    /**
     * Returns a single entity matching the given {@link Qo} or {@link Optional#empty()} if none was found.
     *
     * @param query must not be {@literal null}.
     * @return a single entity matching the given {@link Qo} or {@link Optional#empty()} if none was found.
     * @throws IncorrectResultSizeDataAccessException if the Qo yields more than one
     *                                                result.
     */
    @Override
    public Mono<T> findOne(Qo<T> query) {
        return Mono.justOrEmpty(getQuery(query))
                .flatMap(querying -> mongoOperations.findOne(querying, entityInformation.getJavaType(),
                        entityInformation.getCollectionName()));
    }

    /**
     * Returns all entities matching the given {@link Qo}. In case no match could be found an empty
     * {@link Iterable} is returned.
     *
     * @param query must not be {@literal null}.
     * @return all entities matching the given {@link Qo}.
     */
    @Override
    public Flux<T> findAll(Qo<T> query) {
        return Mono.justOrEmpty(getQuery(query))
                .flatMapMany(querying -> mongoOperations.find(querying,
                        entityInformation.getJavaType(),
                        entityInformation.getCollectionName()))
                .doOnNext(t -> t.setCurrentUser(query.getUser()));
    }

    /**
     * Returns all entities matching the given {@link Qo} applying the given {@link Sort}. In case no match could
     * be found an empty {@link Iterable} is returned.
     *
     * @param query must not be {@literal null}.
     * @param sort  the {@link Sort} specification to sort the results by, may be {@link Sort#empty()}, must not be
     *              {@literal null}.
     * @return all entities matching the given {@link Qo}.
     * @since 1.10
     */
    @Override
    public Flux<T> findAll(Qo<T> query, Sort sort) {
        return Mono.justOrEmpty(getQuery(query))
                .flatMapMany(querying -> mongoOperations.find(querying.with(sort),
                        entityInformation.getJavaType(),
                        entityInformation.getCollectionName()));
    }

    /**
     * Returns a {@link Page} of entities matching the given {@link Qo}. In case no match could be found, an empty
     * {@link Page} is returned.
     *
     * @param query    must not be {@literal null}.
     * @param pageable may be {@link Pageable#unpaged()}, must not be {@literal null}.
     * @return a {@link Page} of entities matching the given {@link Qo}.
     */
    @Override
    public Mono<Page<T>> findAll(Qo<T> query, Pageable pageable) {
        return Mono.justOrEmpty(getQuery(query))
                .flatMap(querying -> mongoOperations.find(querying.with(pageable),
                                entityInformation.getJavaType(), entityInformation.getCollectionName())
                        .doOnNext(t -> t.setCurrentUser(query.getUser()))
                        .collectList()
                        .flatMap(list -> ReactivePageableExecutionUtils.getPage(list, pageable, this.count(query))))
                .defaultIfEmpty(Page.empty());
    }

    /**
     * Returns the number of instances matching the given {@link Qo}.
     *
     * @param query the {@link Qo} to count instances for, must not be {@literal null}.
     * @return the number of instances matching the {@link Qo}.
     */
    @Override
    public Mono<Long> count(Qo<T> query) {
        return Mono.justOrEmpty(getQuery(query))
                .flatMap(querying -> this.mongoOperations.count(querying, entityInformation.getJavaType(),
                        entityInformation.getCollectionName()))
                .defaultIfEmpty(0L);
    }

    /**
     * 通过特定键的集合查询
     *
     * @param key    键
     * @param values 集合
     * @return 结果
     */
    @Override
    public Flux<T> findAllByValues(String key, List<?> values) {
        Criteria criteria = Criteria.where(key).in(values);
        Query query = new Query(criteria);
        return mongoOperations.find(query,
                entityInformation.getJavaType(),
                entityInformation.getCollectionName());
    }

    /**
     * Checks whether the data store contains elements that match the given {@link Qo}.
     *
     * @param query the {@link Qo} to use for the existence check, must not be {@literal null}.
     * @return {@literal true} if the data store contains elements that match the given {@link Qo}.
     */
    @Override
    public Mono<Boolean> exists(Qo<T> query) {
        return Mono.justOrEmpty(getQuery(query))
                .flatMap(querying -> mongoOperations.exists(querying,
                        entityInformation.getJavaType(), entityInformation.getCollectionName()))
                .defaultIfEmpty(false);
    }

    /**
     * 删除全部
     *
     * @param qo 查询实体
     * @return 结果
     */
    @Override
    public Mono<Void> deleteAll(Qo<T> qo) {
        return Mono.justOrEmpty(getQuery(qo))
                .flatMap(querying -> mongoOperations.remove(querying, entityInformation.getJavaType(),
                        entityInformation.getCollectionName()))
                .then(Mono.empty());
    }

    private Optional<Query> getQuery(Qo<T> qo) {
        return qo.getQuery(entityInformation);
    }
}
