package com.flyfish.framework.r2dbc.query.converter;

import com.flyfish.framework.context.SpringContext;
import com.flyfish.framework.query.spi.adaptor.CriteriaAdaptor;
import com.flyfish.framework.query.spi.converter.QueryConverter;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.data.domain.Example;
import org.springframework.data.r2dbc.mapping.R2dbcMappingContext;
import org.springframework.data.relational.core.query.Criteria;
import org.springframework.data.relational.core.query.Query;
import org.springframework.data.relational.repository.query.RelationalExampleMapper;
import org.springframework.util.function.SingletonSupplier;

import java.util.Optional;

/**
 * r2dbc查询转换器
 *
 * @author wangyu
 */
@Slf4j
@RequiredArgsConstructor
public class R2dbcQueryConverter implements QueryConverter<Query> {

    private final CriteriaAdaptor<Criteria> adaptor;

    private final SingletonSupplier<RelationalExampleMapper> exampleMapper = SingletonSupplier.of(R2dbcQueryConverter::getExampleMapper);

    private static RelationalExampleMapper getExampleMapper() {
        R2dbcMappingContext context = SpringContext.getBean(R2dbcMappingContext.class);
        return new RelationalExampleMapper(context);
    }

    /**
     * 转换查询
     *
     * @param query 封装的查询
     * @return 本地查询
     */
    @Override
    public Query convert(com.flyfish.framework.query.Query query) {
        Criteria condition = Criteria.where("delete").not(true);
        // 添加条件
        Criteria criteria = getCriteria(query);
        if (null != criteria) {
            condition = condition.and(criteria);
        }
        // 构建查询
        Query result = Query.query(condition);
        if (null != query.getSort()) {
            result = result.sort(query.getSort());
        }
        if (null != query.getPageable()) {
            result = result.with(query.getPageable());
        }
        if (CollectionUtils.isNotEmpty(query.getFields())) {
            result = result.columns(query.getFields());
        }
        return result;
    }

    /**
     * 获取真实条件
     *
     * @param query 查询信息
     * @return 结果
     */
    private Criteria getCriteria(com.flyfish.framework.query.Query query) {
        // 当且仅当有example且正确生成条件时，才允许使用example构造
        if (null != query.getExample()) {
            Optional<Criteria> optional = ofExample(query.getExample());
            if (optional.isPresent()) {
                return optional.get();
            }
        }
        // 如果存在查询定义，直接使用查询定义
        if (null != query.getDefinition()) {
            return query.getDefinition().build(adaptor);
        }
        return null;
    }

    /**
     * 通过example进行匹配
     *
     * @param example 示例对象
     * @return 结果
     */
    private Optional<Criteria> ofExample(Example<?> example) {
        return exampleMapper.obtain().getMappedExample(example).getCriteria()
                .map(cd -> (Criteria) cd);
    }
}
