package com.flyfish.framework.configuration.jwt;

import com.flyfish.framework.utils.ReactiveRedisOperations;
import com.flyfish.framework.utils.UUIDUtils;
import io.jsonwebtoken.*;
import io.jsonwebtoken.io.Decoders;
import io.jsonwebtoken.io.DecodingException;
import io.jsonwebtoken.security.Keys;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.http.HttpHeaders;
import org.springframework.http.server.reactive.ServerHttpRequest;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.userdetails.User;
import org.springframework.web.server.ServerWebExchange;
import reactor.core.publisher.Mono;

import javax.annotation.Resource;
import javax.crypto.SecretKey;
import java.time.Duration;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * token提供者，负责生成token
 *
 * @author wangyu
 */
@Slf4j
public class TokenProvider implements InitializingBean {

    public static final String AUTHORIZATION_HEADER = "Authorization";
    private static final String AUTHORITIES_KEY = "auth";
    private static final String TOKEN_BLOCK_PREFIX = "tk_blk_";
    private final Boolean remember;
    private final long tokenValidityInMilliseconds;
    private final long tokenValidityInMillisecondsForRememberMe;

    @Resource
    private ReactiveRedisOperations reactiveRedisOperations;
    @Resource
    private JwtProperties jwtProperties;

    private SecretKey key;

    public TokenProvider(JwtProperties jwtProperties) {
        this.remember = jwtProperties.isRemember();
        this.tokenValidityInMilliseconds = jwtProperties.getTokenValidityInSeconds() * 1000;
        this.tokenValidityInMillisecondsForRememberMe = jwtProperties.getTokenValidityInSecondsForRememberMe() * 1000;
    }

    @Override
    public void afterPropertiesSet() {
        byte[] keyBytes = Decoders.BASE64.decode(jwtProperties.getBase64Secret());
        this.key = Keys.hmacShaKeyFor(keyBytes);
    }

    /**
     * 从上下文获取token
     *
     * @return 结果
     */
    public Optional<String> retrieveToken(ServerWebExchange exchange) {
        ServerHttpRequest request = exchange.getRequest();
        String bearerToken = request.getHeaders().getFirst(AUTHORIZATION_HEADER);
        if (StringUtils.isNotBlank(bearerToken) && bearerToken.startsWith("Bearer ")) {
            return Optional.of(bearerToken.substring(7));
        }
        // 从cookies里头找
        if (MapUtils.isNotEmpty(request.getCookies())) {
            return request.getCookies().keySet().stream().filter(name -> name.equals(AUTHORIZATION_HEADER))
                    .findFirst().flatMap(name -> request.getCookies().get(name).stream()
                            .filter(value -> StringUtils.isNotBlank(value.getValue()) && value.getValue().startsWith("Bearer-"))
                            .findFirst()
                    ).map(cookie -> cookie.getValue().substring(7));
        }
        return Optional.empty();
    }

    /**
     * 添加token逻辑
     *
     * @param exchange       交换报文
     * @param authentication 认证信息
     */
    public void addToken(ServerWebExchange exchange, Authentication authentication) {
        String token = createToken(authentication, remember);
        HttpHeaders headers = exchange.getResponse().getHeaders();
        // app用户从头部返回，方便获取
        headers.add("Token", token);
        headers.add("Token-Valid-Time", String.valueOf(remember ? tokenValidityInMillisecondsForRememberMe :
                tokenValidityInMilliseconds));
    }

    /**
     * 移除token
     *
     * @param exchange 交换报文
     * @return 结果
     */
    public Mono<Void> removeToken(ServerWebExchange exchange) {
        return Mono.justOrEmpty(retrieveToken(exchange))
                .map(this::parseToken)
                .flatMap(token -> {
                    long distance = token.getExpiration().getTime() - System.currentTimeMillis();
                    String key = getCacheKey(token.getId());
                    return reactiveRedisOperations.set(key, true, Duration.ofMillis(distance)).then();
                })
                .onErrorResume(e -> Mono.empty());
    }

    /**
     * 创建token
     *
     * @param authentication 认证信息
     * @param rememberMe     记住我标记
     * @return 结果
     */
    public String createToken(Authentication authentication, boolean rememberMe) {
        String authorities = authentication.getAuthorities().stream()
                .map(GrantedAuthority::getAuthority)
                .collect(Collectors.joining(","));

        long now = (new Date()).getTime();
        long duration = rememberMe ? this.tokenValidityInMillisecondsForRememberMe : this.tokenValidityInMilliseconds;
        Date validity = new Date(now + duration);

        return Jwts.builder()
                .subject(authentication.getName())
                .id(UUIDUtils.generateShortUuid())
                .claim(AUTHORITIES_KEY, authorities)
                .signWith(key)
                .expiration(validity)
                .compact();
    }

    /**
     * token解析第一步，获取认证。此处通过claims就能知道token是哪个端的
     *
     * @param token token
     * @return 结果
     */
    public Authentication getAuthentication(String token) {
        Claims claims = parseToken(token);
        Collection<? extends GrantedAuthority> authorities =
                Arrays.stream(claims.get(AUTHORITIES_KEY).toString().split(","))
                        .filter(StringUtils::isNotBlank)
                        .map(SimpleGrantedAuthority::new)
                        .collect(Collectors.toList());

        User principal = new User(claims.getSubject(), "", authorities);
        return new UsernamePasswordAuthenticationToken(principal, token, authorities);
    }

    /**
     * 解析token
     *
     * @param token 用户token、
     * @return 结果
     */
    public Claims parseToken(String token) {
        return Jwts.parser()
                .decryptWith(key)
                .build()
                .parseEncryptedClaims(token)
                .getPayload();
    }

    private String getCacheKey(String jti) {
        return TOKEN_BLOCK_PREFIX + jti;
    }

    public Mono<String> validateToken(String authToken) {
        if (StringUtils.isBlank(authToken)) {
            log.debug("no valid JWT token found");
            return Mono.empty();
        }
        try {
            Claims claims = parseToken(authToken);
            return reactiveRedisOperations.hasKey(getCacheKey(claims.getId()))
                    .filter(result -> !result)
                    .thenReturn(claims.getSubject());
        } catch (io.jsonwebtoken.security.SecurityException | MalformedJwtException e) {
            log.info("Invalid JWT signature.");
            log.trace("Invalid JWT signature trace: {}", e, e);
        } catch (ExpiredJwtException e) {
            log.info("Expired JWT token.");
            log.trace("Expired JWT token trace: {}", e, e);
        } catch (UnsupportedJwtException e) {
            log.info("Unsupported JWT token.");
            log.trace("Unsupported JWT token trace: {}", e, e);
        } catch (IllegalArgumentException e) {
            log.info("JWT token compact of handler are invalid.");
            log.trace("JWT token compact of handler are invalid trace: {}", e, e);
        } catch (DecodingException e) {
            log.info("Token解析失败！");
            log.trace("Token解析失败: {}", e, e);
        }
        return Mono.empty();
    }
}
