package com.flyfish.framework.service;

import com.flyfish.framework.bean.SyncVo;
import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.domain.base.DomainService;
import com.flyfish.framework.domain.base.Qo;
import org.springframework.data.domain.Page;

import java.util.Collection;
import java.util.List;
import java.util.Optional;

/**
 * 基础的Service，提供同步数据结构
 * 代理Repo的所有方法，内部封装常用的全部操作
 *
 * @param <T> 实体泛型
 */
public interface BaseService<T extends Domain> extends DomainService {

    /**
     * 查询
     *
     * @param query 查询模型
     * @return 结果
     */
    Optional<T> getOne(Qo<T> query);

    /**
     * 通过Id查询
     *
     * @param id 主键
     * @return 结果
     */
    Optional<T> getById(String id);

    /**
     * 通过实体名称访问
     *
     * @param name 名称
     * @return 结果
     */
    Optional<T> getByName(String name);

    /**
     * 根据ID集合来查询
     *
     * @param ids id列表
     * @return 结果
     */
    List<T> getByIds(Collection<String> ids);

    /**
     * 通过某个字段的集合包含查询
     *
     * @param key    键
     * @param values 值
     * @return 结果
     */
    <K> List<T> getByValues(String key, List<K> values);

    /**
     * 查询列表
     *
     * @param query 查询
     * @return 结果
     */
    List<T> getList(Qo<T> query);

    /**
     * 分页查询列表
     *
     * @param query 查询实体
     * @return 结果
     */
    Page<T> getPageList(Qo<T> query);

    /**
     * 获取所有对象
     *
     * @return 结果
     */
    List<T> getAll();

    /**
     * 查询总记录数
     *
     * @return 总数
     */
    Long countAll();

    /**
     * 查询总记录数
     *
     * @param query 查询实体
     * @return 结果
     */
    Long count(Qo<T> query);

    /**
     * 插入新记录
     *
     * @param entity 实体
     */
    T create(T entity);

    /**
     * 插入 不插入null字段
     *
     * @param entity 实体
     */
    T createSelective(T entity);

    /**
     * 删除
     *
     * @param entity 实体
     */
    void delete(T entity);

    /**
     * 根据Id删除
     *
     * @param id 主键
     */
    void deleteById(String id);

    /**
     * 根据id更新
     *
     * @param entity 实体
     */
    T updateById(T entity);

    /**
     * 不update null
     *
     * @param entity 实体
     */
    T updateSelectiveById(T entity);

    /**
     * 根据ID集合批量删除
     *
     * @param ids id结合
     */
    void deleteBatchByIds(List<String> ids);

    /**
     * 通过某一个键删除所有
     *
     * @param qo 查询实体
     */
    void deleteAll(Qo<T> qo);

    /**
     * 删除全部实体，危险！
     */
    void deleteAll();

    /**
     * 批量更新
     *
     * @param entities 要批量更新的集合
     */
    List<T> updateBatch(List<T> entities);

    /**
     * 同步数据
     *
     * @param entities 数据们
     * @return 结果
     */
    SyncVo<T> sync(List<T> entities);

}
